#!/usr/bin/env python
# -*- coding: UTF-8 -*-

from __future__ import with_statement

"""Call GPSBabel and modify the output

GPSBabel has many option, for far too many file types. I use a limited 
set of formats (kml, kmz, and gpx) for my navigation and sharing. The
first purpose of this tool is to limit the number of options, and treat 
files consistently for specific tasks:

* Import a gpx file from my Garmin and store as km[lz] for use with 
  Google Earth. This means that parts of the track have to be glued 
  together.
* Import a gpx file from the "Fietsrouteplanner" or Bobcat to km[lz]. 
  In this case the tracks should not be glued.
* Export to gpx for use with my Garmin. This means limiting the number 
  of points to 500, the maximum my Garmin will handle in a single track 
  without truncating.
  
Since I don't like the standard kml output of GPSBabel, I have to post 
process the (xml) output. When producing km[lz], I also calculate the 
length of the track, since Google Earth has no easy interface to do so.

Normally when translating kml to kml, the descriptions are lost. 
I copy them over.

Whereabouts GPS handling tool
Copyright © 2008  Maarten Sneep

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

__author__ = "Maarten Sneep (maarten.sneep@xs4all.nl)"
__version__ = "1.5"
__date__ = "15-05-2009"
__copyright__ = "Copyright © 2008, 2009 Maarten Sneep"
__license__ = "GPL version 2 or later."


import os
import sys
import re
import random
import tempfile
import zipfile
import datetime
import math
import re
from xml.dom import minidom
from optparse import OptionParser

def Main():
    """Main program "Whereabouts". """
    scriptname = os.path.splitext(os.path.basename(sys.argv[0]))[0].lower()
    (options, args) = cmdline()
    verbose = options.verbose
    
    # The output file type is determined by the name of the script.
    # If the default script has an alias to it with a specific name, 
    # some options are set by default. If we have a different name,
    # we check some options, and assigne the wanted script name.
    if scriptname not in ["makegpx", "makekmz", "makekml", "makegdb"]:
        if options.makegpx:
            scriptname = "makegpx"
        elif options.makegdb:
            scriptname = "makegdb"
        else:
            if options.compress: 
                scriptname = "makekmz"
            else: 
                scriptname = "makekml"
    
    if options.device_upload:
        outtype="usb"
        
    else:
        outtype = scriptname[4:]
    if verbose: print("Main()\n\tproducing %s" % (outtype,))
    
    if options.device_download:
        args.append("device.usb")
    
    if options.combine:
        # if track-parts will be combined, then we must process each input file in turn.
        files = []
        for item in args:
            if len(options.outdir) == 0:
                basedir = os.path.dirname(item)
            else:
                basedir = options.outdir
            
            outfile = os.path.join(basedir, os.path.splitext(item)[0] + "." + outtype)
            idx = 1
            while os.path.isfile(outfile) or outfile in files:
                outfile = os.path.join(basedir, 
                    os.path.splitext(item)[0] + "-" + str(idx) + "." + outtype)
                idx += 1
            files.append(([item],outfile))
    else:
        # Process all files in one go, as the tracks themselves will not be glued together.
        if len(options.outdir) == 0:
            basedir = os.path.dirname(args[0])
        else:
            basedir = options.outdir
        
        outfile = os.path.join(basedir, os.path.splitext(args[0])[0] + "." + outtype)
        idx = 1
        while os.path.isfile(outfile):
            outfile = os.path.join(basedir, 
                os.path.splitext(args[0])[0] + "-" + str(idx) + "." + outtype)
            idx += 1
        files = [(args, outfile)]
    
    # loop over the items and output files (can be a single item).
    for items,outfile in files:
        # The converter object handles the conversion and builds the command line.
        converter = Converter(options.tool, verbose=options.verbose)
        
        # Produce a list of readers and a set of track descriptions descriptions
        readers = []
        descriptions = {}
        for item in items:
            # get the extension of the input file, and determine the 
            # type of the input file from it.
            extension = (os.path.splitext(item)[1].lower())[1:]
            
            # skip file if the type is unknown.
            if extension not in ['gpx', 'kml', 'kmz', 'gdb', 'usb', 'xml']: continue
            
            # create a reader object, which has a type determined by the extension.
            # The initializer for these objects expect a filename and a verbose flag.
            # (actually there are more named parameters, but we can ignore those for
            # the reader object).
            reader = instanceFromName(extension, item, verbose=verbose)
            converter.addinputfile(reader)
            
            # The act of adding the reader as an input file to the converter, has
            # the reader gather the tack names and the descriptions from the input file
            # (only for kml/kmz files). Put all those combinations in a single dictionary.
            if reader.descriptions: 
                for name, descr in reader.descriptions.items(): descriptions[name] = descr
                
            # Since kmz files are expanded into a temporary file when the reader object is 
            # initialized, we must keep these objects alive until after gpsbabel has done its job.
            # (The temporary files are removed in the __del__() method).
            readers.append(reader)
        
        # if we have no readers, then we commit suicide here
        if not readers: sys.exit(1)
        
        # delete all waypoints if requested to do so.
        if options.delete_waypoints:
            converter.addfilter("nuketypes,waypoints")

        # We alway filter the tracks for points closer than 2 meter apart. 
        # With an accuracy of typically > 8 m, keeping more is pointless.
        converter.addfilter("position,distance=2m")
        
        # if options.length is larger than zero, it is set by the used, and we limit the length 
        # of the track to this number of points.
        if options.length > 0:
            converter.addfilter("simplify,count=%(length)d" % {'length':options.length})
        
        # Add a filter option to combine the track parts if requested.
        if options.combine:
            converter.addfilter("track,pack,split=1d,title=%Y-%m-%d ")
        
        # Create a writer instance (same object as a reader, just used slightly differently).
        # The descriptions gathered earlier are passed on here. The writer object is added 
        # to the converter object.
        writer = instanceFromName(outtype, outfile, 
            verbose=options.verbose, inputs=items, descriptions=descriptions)
        converter.addoutputfile(writer)
        
        # always expect the unexpected, catch exceptions.
        try:
            converter.execute()
        except ToolException, e:
            print("GPSBabel returned an error, exit value: %s" % (str(e),))
            sys.exit(e.value)
        finally:
            # This is an interesting one. A few things, especially on kml output files,
            # can only be done at the end: kml rewriting and cleanup, track length 
            # calculation and compression. This is done in the __del__() method.
            del writer
            del readers
        
        # This one is for use in Automator. 
        # By printing the produced filenames to stdout, Automator can continue with those files.
        if options.tool: sys.stdout.write("'%s' " %(outfile,))
    if options.tool: sys.stdout.write("\n")
    
def distance(loc1, loc2):
    """Calculate geodesic distance (in m) between two points specified by
latitude/longitude (in numeric degrees) using Vincenty inverse formula 
for ellipsoids. 

Arguments are tuples of (lat, lon) pairs. 

Code based on a Javascript version by © 2002-2008 Chris Veness. 

That code contains the following: "You are welcome to re-use these scripts 
[without any warranty express or implied] provided you retain my copyright notice 
and when possible a link to my website (under a LGPL license)."

The originating web page: http://www.movable-type.co.uk/scripts/latlong-vincenty.html
"""
    a = 6378137 # in m
    b = 6356752.3142 # in m
    f = 1/298.257223563
    deg2rad = math.pi / 180
    lat1, lon1 = (loc1[0] * deg2rad, loc1[1] * deg2rad)
    lat2, lon2 = (loc2[0] * deg2rad, loc2[1] * deg2rad)
    L = (lon2 - lon1)
    U1 = math.atan((1-f) * math.tan(lat1))
    U2 = math.atan((1-f) * math.tan(lat2))
    sinU1 = math.sin(U1)
    cosU1 = math.cos(U1)
    sinU2 = math.sin(U2)
    cosU2 = math.cos(U2)
    Lambda = L
    LambdaP = 2*math.pi
    iterLimit = 20
    
    while math.fabs(Lambda-LambdaP) > 1e-12 and iterLimit > 0:
        sinLambda = math.sin(Lambda)
        cosLambda = math.cos(Lambda)
        sinSigma = math.sqrt((cosU2*sinLambda)**2 + (cosU1*sinU2-sinU1*cosU2*cosLambda)**2)
        if (sinSigma==0): return 0 # co-incident points
        cosSigma = sinU1*sinU2 + cosU1*cosU2*cosLambda
        sigma = math.atan2(sinSigma, cosSigma)
        # sinSigma ≠ 0, already caught by condition above.
        sinAlpha = cosU1 * cosU2 * sinLambda / sinSigma
        cosSqAlpha = 1 - sinAlpha**2
        try:
            cos2SigmaM = cosSigma - 2*sinU1*sinU2/cosSqAlpha
        except:
            cos2SigmaM = 0 # equatorial line: cosSqAlpha=0
        C = f/16*cosSqAlpha*(4+f*(4-3*cosSqAlpha));
        LambdaP = Lambda;
        Lambda = L + (1-C) * f * sinAlpha * (sigma + C*sinSigma*(cos2SigmaM+C*cosSigma*(-1+2*cos2SigmaM**2)))
        iterLimit -= 1
    
    if iterLimit == 0:
        return None  # formula failed to converge
    
    uSq = cosSqAlpha * (a**2 - b**2) / (b**2)
    A = 1 + uSq/16384*(4096+uSq*(-768+uSq*(320-175*uSq)))
    B = uSq/1024 * (256+uSq*(-128+uSq*(74-47*uSq)))
    deltaSigma = B*sinSigma*(cos2SigmaM+B/4*(cosSigma*(-1+2*cos2SigmaM**2) - B/6*cos2SigmaM*(-3+4*sinSigma**2)*(-3+4*cos2SigmaM**2)))
    return b*A*(sigma-deltaSigma);

def totalDistance(lats, lons):
    """Calculate the total distance of a track"""
    Tdistance = 0.0
    for ii in range(len(lats) - 1):
        d = distance( (lats[ii], lons[ii]), (lats[ii+1], lons[ii+1]) )
        if d: Tdistance += d
    return Tdistance

def distanceFromPlacemark(placemark):
    """Extract the information in a "Placemark" DOM object, 
and calculate the length of that placemark.

The contents of the 'coordinates' DOM object are not exactly xml."""
    trackData = placemark.getElementsByTagName('coordinates')[0].firstChild.data
    points = trackData.split() # split on whitespace
    lats = []
    lons = []
    for point in points:
        coords = point.split(',') # split on comma, coords are longitude,latitude,altitude.
        lats.append(float(coords[1]))
        lons.append(float(coords[0]))
    # dump data in totalDistance()
    return totalDistance(lats, lons) 

class ToolException(Exception):
    """Simple exception to capture errors of the command line tool"""
    def __init__(self,ecode):
        self.value = ecode
    def __str__(self):
        return repr(self.value)
    
class Converter(object):
    """Gather the arguments for gpsbabel, and call the tool.
    
The gpsbabel tool must be present in the $PATH environment variable. 
Of course, as Windows® likes to do things differently, this won't 
work there. Supply the full path in that case to the appropriate 
command-line option.
    """
    def __init__(self, tool, verbose=None):
        # if tool is given, we assume a full path, and store the appropriate parts.
        if len(tool) > 0:
            self.toolpath = os.path.dirname(tool)
            self.toolname = os.path.basename(tool)
        else:
            self.toolpath = ""
            self.toolname = "gpsbabel"
        # first argument is the name of the tool.
        self.args = [self.toolname]
        self.verbose = verbose
        if self.verbose:
            print("Converter::__init__")
            print("\tself.args: %s\n\tself.toolpath: %s\n\tself.toolname: %s" % (self.args, self.toolpath,self.toolname))
    def addinputfile(self, obj):
        if self.verbose:
            print("Converter::addinputfile")
        self.args.extend(obj.read())
    def addoutputfile(self, obj):
        if self.verbose:
            print("Converter::addoutputfile")
        self.args.extend(obj.write())
    def addfilter(self, str):
        if self.verbose:
            print("Converter::addfilter")
        self.args.extend(["-x", str])
    def execute(self):
        if self.verbose:
            sys.stdout.write("Converter::execute\n\t")
            for item in self.args:
                sys.stdout.write("%s " %(item,))
            sys.stdout.write("\n")
        if self.toolpath:
            # if toolpath was given, we use it. This also works on Windows.
            ecode = os.spawnv(os.P_WAIT, os.path.join(self.toolpath, self.toolname), self.args)
        else:
            # if toolpath is empty, we use the version which searches 
            # the $PATH environment variable (not available on Windows).
            ecode = os.spawnvp(os.P_WAIT, "gpsbabel", self.args)
        if ecode != 0:
            raise ToolException(ecode)

class kml(object):
    file = ""
    def __init__(self,file, verbose=None, inputs=None, descriptions=None):
        self.file = file
        self.verbose = verbose
        self.inputs = inputs
        if self.verbose:
            print("kml::__init__")
            print("\tfile: %s" % (self.file,))
        self.descriptions = descriptions
    def read(self):
        """Add the file as an input file. Extract the descriptions. Returns the appropriate command-line option for GPSBabel."""
        self.writer = False
        answer = ["-i", "kml", 
            "-f", self.file,
            "-x", "transform,trk=rte,del"]
        if self.verbose:
            print("kml::read")
            print("\t%s" % (" ".join(answer),))
        self.descriptions = self.extract_descriptions()
        return answer
    def write(self):
        """Add self.file as an output file. Returns the appropriate command-line option for GPSBabel."""
        self.writer = True
        seed = os.urandom(128)
        if self.verbose: print("\tseed: %d" % hash(seed))
        random.seed(seed)
        randomvalues = {'width': random.uniform(2.5, 4.5),
            'alpha': random.randint(230, 255),
            'r': random.randint(0, 255),
            'g': random.randint(0, 255), 
            'b': random.randint(0, 255)}
        answer = ["-o", 
            "kml,lines=1,points=0,line_width=%(width).1f,floating=0,extrude=0,trackdata=0,trackdirection=0,units=m,labels=0,line_color=%(alpha)02x%(b)02x%(g)02x%(r)02x" % randomvalues , 
            "-F", self.file]
        if self.verbose:
            print("kml::write")
            print("\t%s" % (" ".join(answer),))
        return answer
        
    def kml_cleanup(self):
        """Read the (GPSBabel) produced kml, pick out the wanted elements, and write the simplified data out again.

This also adds the track lengths and appends those to the descriptions. Oh, and it tries to use the 
saved descriptions as well."""
        # This code really should be split into smaller pieces, but it really is fairly straightforward.
        if not self.writer: raise IOError, "Only cleanup on writing the file, not when reading the file."
        if self.verbose: print("kml::cleanup")
        # read and parse the kml
        doc = minidom.parse(self.file)
        # prepare for writing a new document
        impl = minidom.getDOMImplementation()
        newdoc = impl.createDocument(None,"kml",None)
        
        # get all Folders, and find the "Tracks" folder.
        folders = [e 
            for e in doc.getElementsByTagName('Folder') 
            if e.nodeType == e.ELEMENT_NODE]
        names = [folder.getElementsByTagName('name')[0].firstChild.data for folder in folders]
        trackFolder = [e 
            for e in folders[names.index(u'Tracks')].getElementsByTagName('Folder') 
            if e.nodeType == e.ELEMENT_NODE]
        
        # get the names of the tracks.
        trackNames = []
        for folder in trackFolder:
            nameElement = folder.getElementsByTagName('name')[0]
            name = nameElement.firstChild.data
            if nameElement.nodeType == nameElement.ELEMENT_NODE and name != u'Path':
                trackNames.append(name)
            
        # get the placemarks (the tracks themselves).
        placemarks = [folder.getElementsByTagName('Placemark')[0] for folder in trackFolder]
        
        if self.verbose: print("\tFound %d tracks" % (len(placemarks),))
        
        # if not all tracks were named (can happen with 'Fietsrouteplanner' tracks), 
        # construct new names based on the input files.
        if len(trackNames) < len(placemarks):
            trackNames = [os.path.splitext(os.path.basename(e))[0] 
                for e in self.inputs][0:len(placemarks)]
        
        if self.verbose: print("\tCurrent track names: '%s'" % ("', '".join(trackNames),))
        
        # find line style names used in the tracks.
        styleNames = []
        for item in placemarks:
            styleName = item.getElementsByTagName('styleUrl')[0].firstChild.data[1:]
            if styleName not in styleNames:
                styleNames.append(styleName)
                if self.verbose:
                    print("\tSelected style '%s'" % (styleName,))
        
        # filter the lines styles we actually need
        styles = [item for item in doc.getElementsByTagName('Style') 
            if 'id' in item.attributes.keys() 
                and item.attributes['id'].value in styleNames]
        
        # build the new doc
        for item in doc.documentElement.attributes.values():
            newdoc.documentElement.setAttribute(item.name, item.value) 
        
        # The master element within the root element.
        docElement = newdoc.documentElement.appendChild(newdoc.createElement("Document"))
        
        # make sure the folders are "open" (i.e. contents visible in Google Earth).
        openFolder = docElement.appendChild(newdoc.createElement("open"))
        openFolder.appendChild(newdoc.createTextNode("1"))
        
        # if we have less line styles than tracks, we have to make up more as we want each to have a unique color.
        if len(styles) < len(placemarks):
            baseStyle = styles[0]
            baseStyleName = baseStyle.attributes['id'].value
            styles = []
            styleNames = []
            for placemark in placemarks:
                newStyle = newdoc.createElement("Style")
                newStyle.setAttribute("id", baseStyleName + str(len(styles) + 1))
                newLineStyle = newdoc.createElement("LineStyle")
                newColor = newdoc.createElement("color")
                # create a random color
                values = {
                    'alpha': random.randint(230, 255),
                    'r': random.randint(0, 255),
                    'g': random.randint(0, 255), 
                    'b': random.randint(0, 255)}
                newColor.appendChild(newdoc.createTextNode("%(alpha)02x%(b)02x%(g)02x%(r)02x" % values))
                if self.verbose: print("\tnew colour: a%(alpha)02x b%(b)02x g%(g)02x r%(r)02x" % values)
                newWidth = newdoc.createElement("width")
                # create a random width
                newWidth.appendChild(newdoc.createTextNode("%3.1f" % (random.random()*2.0 + 2.5,)))
                newLineStyle.appendChild(newColor)
                newLineStyle.appendChild(newWidth)
                newStyle.appendChild(newLineStyle)
                styles.append(newStyle)
                styleNames.append(newStyle.attributes["id"].value)
                
                # make sure the track refers to the new style.
                newStyleUrl = newdoc.createElement("styleUrl")
                newStyleUrl.appendChild(newdoc.createTextNode("#" + newStyle.attributes["id"].value ))
                
                oldStyleUrl = placemark.getElementsByTagName('styleUrl')[0]
                placemark.replaceChild(newStyleUrl, oldStyleUrl)
        
        # add all line styles to the new document
        for style in styles: docElement.appendChild(style)
        
        # compile re to detect length in previous description.
        descriptionfilter = re.compile(r"(.*)((<p>De route is \d*\.\d&nbsp;km lang\.</p>)|(<p>Track is \d*\.\d&nbsp;km long\.</p>))$", re.IGNORECASE | re.DOTALL | re.UNICODE )
        
        # loop over all tracks.
        for ii in range(len(placemarks)):
            newNameElement = newdoc.createElement("name")
            newNameElement.appendChild(newdoc.createTextNode(trackNames[ii]))
            
            # replace the existing name with the new one. The existing one 
            # is probably 'Path', we want something more descriptive.
            try:
                currentName = placemarks[ii].getElementsByTagName('name')[0]
            except:
                currentName = None
            if currentName:
                placemarks[ii].replaceChild(newNameElement, currentName)
            else:
                placemarks[ii].appendChild(newNameElement)
            
            # calculate length, and make an appropriate description
            # (carry the old one over if it existed)
            length = distanceFromPlacemark(placemarks[ii])
            if self.verbose:
                print("\tTrack '%s' has a length of %.1f km" % (trackNames[ii], length/1000))
            
            newDescription = "<p>Route is %.1f&nbsp;km lang.</p>" % (length/1000.,)
            search_result = None
            
            try:
                description = self.descriptions[trackNames[ii]]
                if self.verbose:
                    print("\tFound previous description for '%s': %s" % (trackNames[ii],description))
            except:
                if self.verbose:
                    print("\tNo previous description for '%s'" % (trackNames[ii],))
                description = None
            
            if description is not None:
                try:
                    search_result = descriptionfilter.search(description)
                except:
                    if self.verbose:
                        print("\t(Ir)Regular expression.")
                    raise
                
                if self.verbose: 
                    print("\tSearch result: %s" %(str(search_result),))
                
                if search_result is not None:
                    newDescription = "%s\n<p>Route is %.1f&nbsp;km lang.</p>" % (search_result.group(1),length/1000.)
                else:
                    newDescription = "%s\n<p>Route is %.1f&nbsp;km lang.</p>" % (description,length/1000.)
            if self.verbose:
                print("\tDescription for '%s': '%s'" % (trackNames[ii],newDescription))

            descr = newdoc.createElement("description")
            descr.appendChild(newdoc.createTextNode(newDescription))
            
            # we already know that there are no descriptions in the GPSBabel processed
            # file, so we can safely append.
            placemarks[ii].appendChild(descr)
            newSnippet = newdoc.createElement("Snippet")
            newSnippet.setAttribute("maxLines", "0")
            placemarks[ii].appendChild(newSnippet)
            
            docElement.appendChild(placemarks[ii])
        
        newdoc.normalize()
        if self.verbose: 
            f = open(os.path.splitext(self.file)[0] + "-filtered.kml", "w" )
        else:
            f = open(self.file, "w")
        f.write(newdoc.toxml(encoding="utf-8"))
        f.close()
        
        doc.unlink()
        newdoc.unlink()
        
    def extract_descriptions(self):
        descriptions = {}
        doc = minidom.parse(self.file)
        placemarks = doc.getElementsByTagName('Placemark')
        missingNameIdx = 0
        for placemark in placemarks:
            try:
                name = placemark.getElementsByTagName('name')[0].firstChild.data
            except:
                if missingNameIdx == 0:
                    name = os.path.splitext(os.path.basename(self.file))[0]
                else:
                    name = "%s - %d" % (os.path.splitext(os.path.basename(self.file))[0],missingNameIdx)
                missingNameIdx += 1
            try:
                description = placemark.getElementsByTagName('description')[0].firstChild.data
            except:
                description = None
            descriptions[name] = description
        return descriptions
    
    def __del__(self):
        if self.verbose: print("kml::__del__")
        if self.writer:
            self.kml_cleanup()
        
class gpx(object):
    file = ""
    def __init__(self, file, verbose=None, inputs=None, descriptions=None):
        self.file = file
        self.verbose = verbose
        self.inputs = inputs
        self.descriptions = descriptions
    def read(self):
        return ["-i", "gpx", 
            "-f", self.file, 
            "-x", "transform,trk=rte,del"]
    def write(self):
        return ["-o", "gpx,snlen=12,suppresswhite=1,gpxver=1.1", 
            "-F", self.file]

class xml(gpx):
    pass

class kmz(kml):
    def __init__(self,file, verbose=None, inputs=None, descriptions=None):
        self.realfile = file
        self.file = os.path.join(os.path.dirname(file), os.path.basename(file).split(".")[0] + ".kml")
        self.verbose = verbose
        self.inputs = inputs
        self.descriptions = descriptions
    def read(self):
        try:
            inhandle = zipfile.ZipFile(self.realfile, "r")
            outhandle = open(self.file, "w+b")
            outhandle.write(inhandle.read("doc.kml"))
        finally:
            inhandle.close()
            outhandle.close()
        
        return super(self.__class__,self).read()
    
    def write(self):
        return super(self.__class__,self).write()
        
    def __del__(self):
        if self.writer:
            self.kml_cleanup()
            try:
                outhandle = zipfile.ZipFile(self.realfile, "w", zipfile.ZIP_DEFLATED)
                outhandle.write(self.file, "doc.kml", zipfile.ZIP_DEFLATED)
            finally:
                outhandle.close()
        os.unlink(self.file)

class gdb(object):
    file=""
    def __init__(self, file, verbose=None, inputs=None, descriptions=None):
        self.file = file
        self.verbose = verbose
        self.inputs = inputs
        self.descriptions = descriptions
    def read(self):
        return ["-i", "gdb", 
            "-f", self.file, 
            "-x", "transform,trk=rte,del"]
    def write(self):
        return ["-o", "gdb", 
            "-F", self.file]

class usb(object):
    file="usb:"
    def __init__(self, file, verbose=None, inputs=None, descriptions=None):
        self.verbose = verbose
        self.inputs = inputs
        self.descriptions = descriptions
    def read(self):
        return ["-t", "-w", "-r", 
            "-i", "garmin", 
            "-f", self.file, 
            "-x", "transform,trk=rte,del"]
    def write(self):
        return ["-t", "-w", "-r", "-o", "garmin,snlen=12,snwhite=0", "-F", self.file]

def cmdline():
    scriptname = os.path.basename(sys.argv[0]).split(".")[0].lower()
    usage=u"""usage: %prog [options] infile [infile] [...]
purpose: call gpsbabel in a way that I like its output.

The name of the output file is the same as the input file, with a changed 
extension, and added numbers as needed to prevent overwriting existing 
files. %prog will never overwrite existing files.

Copyright © 2008, 2009  Maarten Sneep

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA."""
    parser = OptionParser(usage=usage, version="%prog "+__version__+" ("+__date__+")")
    parser.add_option("-V", "--verbose", action="store_true", dest="verbose",
        default=False, help="Be chatty")
    parser.add_option("-o", "--outdir", dest="outdir",
        default="", metavar="DIR", 
        help="Put output in DIR, the default is to put the output with the input file.")
    parser.add_option("", "--gpsbabel", dest="tool",
        default="", metavar="TOOL",
        help="Use gpsbabel located in the full path TOOL.")
    parser.add_option("", "--combine", action="store_true",
        dest="combine", default=False,
        help="Concatenate all track-parts into a single track")
    parser.add_option("", "--length", action="store",
        type="int", dest="length", default=0,
        help="Limit the length of the track(s)")
    parser.add_option("","--delete-waypoints", action="store_true",
        dest="delete_waypoints", default=False,
        help="Delete any waypoints in the source file")
    if scriptname not in ["makegpx", "makekmz", "makekml", "makegdb"]:
        parser.add_option("", "--gpx", action="store_true", 
            dest="makegpx", default=True, 
            help="Create gpx from input.")
        parser.add_option("", "--kml", action="store_false", 
            dest="makegpx", default=True, 
            help="Create kml from input.")
        parser.add_option("", "--compress", action="store_true", 
            dest="compress", default=False, 
            help="Compress kml output, making kmz.")
        parser.add_option("", "--gdb", action="store_true",
            dest="makegdb", default=False,
            help="Make MapSource GDB.")
    parser.add_option("", "--device-upload", action="store_true",
        dest="device_upload", default=False, 
        help="Send the tracks and waypoints to the connected usb GPS-device from Garmin")
    parser.add_option("", "--device-download", action="store_true",
        dest="device_download", default=False, 
        help="Retrieve the tracks and waypoints from the connected usb GPS-device from Garmin")
    return parser.parse_args()
    
def instanceFromName(name,*args,**kwargs):
    if name[0] == ".": 
        name = name[1:]
    try:
        classObject = eval(name)
    except NameError, ex:
        sys.stderr.write('Error: "%s"\n' % ex.args)
        raise
    except:
        sys.stderr.write("An error occured during the creation of a '%s' object\n"
                         "Error: %s\n" % (name, sys.exc_info()[0]) )
        raise
    try:
        instance = classObject(*args,**kwargs)
    except:
        sys.stderr.write("An error occured during the creation of a '%s' object\n"
                         "Error: %s\n" % (name, sys.exc_info()[0]) )
        raise
    
    return instance

# Run this code when called directly instead of loaded as a module.
if __name__ == "__main__":
    Main()
